/*
 * Copyright (C) 2004 Javier Ruiz Hidalgo <jrhbcn@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

using Gdk;
using Gtk;

using System;
using System.Diagnostics;
using System.Collections;

using Mono.Unix;


public class CDView  : TreeView
{
	private string sorting;
	private int findkey = -1;
    private CDViewPopup cdview_popup;
    private bool selection_event = true;
    
    /* Events to trigger */
	public delegate void DeleteCDsEventHandler(ArrayList cds_sel);
	public event DeleteCDsEventHandler DeleteCDsEvent;
	
	public delegate void EditCDEventHandler(int key);
	public event EditCDEventHandler EditCDEvent;
	
	public delegate void SelectionChangeEventHandler(ArrayList cds_sel);
	public event SelectionChangeEventHandler SelectionChangeEvent;
    
    
	
	public CDView(string sort) : base(new ListStore (	typeof(int),	/* key */
														typeof(Pixbuf),	/* icon */
														typeof(string), /* name */
														typeof(string)	/* date */
													 ))
	{
				
		RulesHint = true;
        HeadersVisible = false;
        AppendColumn (Catalog.GetString ("Icon"), new CellRendererPixbuf(), "pixbuf", 1);
        AppendColumn (Catalog.GetString ("Name"), new CellRendererText(), "markup", 2);
        //AppendColumn ("Date", new CellRendererText(), "text", 3);
        
		/* Allow multiple selections */
		Selection.Mode = SelectionMode.Multiple;

		EnableSearch = true;
		SearchColumn = 2;

        /* Set sorting of the cdview */
        set_sorting(sort);

		/* Create Popup Menu */
		cdview_popup = new CDViewPopup(sort);
		cdview_popup.CDViewPopupEvent += new CDViewPopup.CDViewPopupEventHandler (on_popup_event);

		/* Select event handlers */			
		KeyPressEvent += new KeyPressEventHandler (on_keypress);
		Selection.Changed += new EventHandler (on_selection_change);
		PopupMenu += new PopupMenuHandler (on_popup);

		search_cds ("");
				
	}
		
	public void stop_selection()
	{
		selection_event = false;
	}

	public void resume_selection()
	{
		selection_event = true;
	}

	public void set_sorting (string sort)
	{
        sorting = sort;
        if (sorting=="name") {
        	(Model as ListStore).SetSortColumnId( 2, Gtk.SortType.Ascending);
        } else {
        	(Model as ListStore).SetSortColumnId( 3, Gtk.SortType.Ascending);
        }
	
	}
	
	public void search_cds (string text)
	{
		/* Clear view */
		(Model as ListStore).Clear();
	
		/* Do the Search on the treeview */
		foreach (CDItem cd in CDCollect.db.cds.Values) {
			if ( cd.Search(text) ) {
				append_cd (cd);
			}
		}
	}
	
	private void append_cd (CDItem cd)
	{
		(Model as ListStore).AppendValues (
			cd.Key,
			cd.Icon.pixbuf(), 
			"<b>" + cd.Name + "</b>\n<small>" + cd.Date.ToShortDateString() + "</small>", 
			cd.Date.ToString("yyyyMMdd") + "-" + cd.Key); 
	}
	
	
	
	public bool select_cd (int key)
	{
    	/* Find the cd on the list */
    	findkey = key;
    	Model.Foreach (new TreeModelForeachFunc(findcd) );
    	
    	/* if not found then return false */
		return (false);
	}
	private bool findcd (TreeModel model, TreePath path, TreeIter iter)
	{
		int k = (int) Model.GetValue (iter, 0);
		if (findkey==k) {
			Selection.SelectPath (path);
			ScrollToCell (path, GetColumn(0), false, 0.5F, 0.5F);
			return(true);
		}
		return(false);
	}	
	
		
	public ArrayList get_selected_cds()
	{
		TreeModel m;
		TreeIter i;
		ArrayList cds_sel = new ArrayList ();
		int key;
	 
		TreePath [] p = Selection.GetSelectedRows (out m);       
        				       
        for (int j=0; j<p.Length; j++) {					
			if ( Selection.TreeView.Model.GetIter (out i, p[j]) ) {
				key = (int) m.GetValue(i, 0);
				cds_sel.Add ((CDItem)CDCollect.db.cds[key]);
			}
		}
       
		return (cds_sel);
	}
	
	
	
	
	
	public void on_keypress (object o, KeyPressEventArgs args)
    {
   								
		/* Delete on supr key */
		if ( args.Event.Key == Gdk.Key.Delete ) { 				
			/* Get selected cds */
			ArrayList cds_sel  = get_selected_cds();			

			/* Launch event to delete selected cds */
			if (cds_sel.Count>0) {
				DeleteCDsEvent (cds_sel);
			}
					
		}
	}

	public void on_selection_change (object o, EventArgs args)
    {
    	if (selection_event) {
			ArrayList cds_sel = get_selected_cds ();
			SelectionChangeEvent (cds_sel);
		}
	}

	protected override bool OnButtonPressEvent(Gdk.EventButton e)
	{
		base.OnButtonPressEvent(e);
		if (e!=null) {
			if (e.Button==3) {
				ArrayList cds_sel = get_selected_cds ();
				cdview_popup.Popup ( cds_sel.Count, e.Button, e.Time);
				return(true);
			}
		}		
		return(true);
	}

	public void on_popup (object o, PopupMenuArgs args)
	{
		ArrayList cds_sel = get_selected_cds ();
		cdview_popup.Popup ( cds_sel.Count, 0, 0);
	}
		
    public void on_popup_event (string action)
	{
		ArrayList cds_sel;
		
		switch (action.ToLower()) {
		
			case "delete":
				/* Get selected cds */
				cds_sel = get_selected_cds();
				
				/* Launch event to delete selected cds */
				if (cds_sel.Count>0) { 				
					DeleteCDsEvent (cds_sel);
				}
				break;
	       
			case "edit":
				/* Make sure there is only one selected cd */
				cds_sel = get_selected_cds();
				
				if (cds_sel.Count!=1) {
					Console.WriteLine (Catalog.GetString ("ERROR: CDView Popup menu selects 'edit' on more than one cd"));
				} else { 
					EditCDEvent (((CDItem)cds_sel[0]).Key);
				}
				break;
			
			case "sortdate":				
				CDCollect.conf.setkey ("prefs/sort", "date");
				set_sorting("date");
				break;
			
			case "sortname":
				CDCollect.conf.setkey ("prefs/sort", "name");
				set_sorting("name");
				break;
		}
	}
	
}

