/*
 * Copyright (C) 2004 Javier Ruiz Hidalgo <jrhbcn@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

using System;
using System.Collections;
using System.Diagnostics;
using System.IO;
using System.Xml;
using System.Text.RegularExpressions;

using Mono.Unix;

using Gtk;
using Gdk;
using Gnome.Vfs;

/*

    import: read from args or external and put things in DB and hashtables
    delete: delete contents from DB and hashtables
    update: update contents in DB and hashtables
    
    read : Read contents from DB to args
    write: Write contents from args to DB
    
*/

public class CDDatabase
{

	/* Public variables */
	public Hashtable cds;
	public Hashtable infoplugins;
	public Hashtable compplugins;
	
	/* Use the sql implementation */
	public static CDDatabaseSql sql;
		

	/* Private variables */
	//private static InfoMimeType infomime;

	/* CONSTRUCTOR */
	public CDDatabase(string dbname,string thumbsdir)
	{

        /* Create/Read the database in disk */
        Trace.WriteLine ("Creating/Opening SQL Database on disk.");
        sql = new CDDatabaseSql (dbname,thumbsdir);	

		/* Create local variables */
		//infomime = new InfoMimeType();

		/* Create hashtables */
		Trace.WriteLine ("Reading plugins from Database.");
		infoplugins = sql.read_infos ();        
		compplugins = sql.read_comps ();

		Trace.WriteLine ("Reading cds from Database.");
   		cds = sql.read_cds ();
        
	}
	
	public int Nfiles ()
	{
		return (sql.get_total_files ());
	}
	
	/* MEMBERS */

	public void import_infoplugin (InfoPluginItem p) 
	{
		/* Delete if it exists */
		if ( infoplugins.Contains (p.Mime) ) {
			delete_infoplugin (p);
		}

		/* Add plugin to Hashtable */
		infoplugins.Add(p.Mime, p);
		
		/* Add plugin to DB */
		sql.write_info (p);
		
	}
	public void import_compplugin (CompPluginItem c) 
	{
		/* Delete if it exists */
		if ( compplugins.Contains (c.Mime) ) {
			delete_compplugin (c);
		}

		/* Add plugin to Hashtable */
		compplugins.Add(c.Mime, c);
		
		/* Add plugin to DB */
		sql.write_comp (c);
	}


	public void delete_infoplugin (InfoPluginItem p)
	{
		delete_infoplugin (p.Mime);
	}	
	public void delete_infoplugin (string mime)
	{
		/* Delete it from the hashtable */
		if (infoplugins.Contains(mime)) {
		  infoplugins.Remove (mime);
		}
		
		/* Remove it from the DB */
		sql.delete_info (mime);
	}	
	public void delete_compplugin (CompPluginItem c)
	{
		delete_compplugin (c.Mime);
	}	

	public void delete_compplugin (string mime)
	{
		/* Delete it from the hashtable */
		if (compplugins.Contains(mime)) {
		  compplugins.Remove (mime);
		}
		
		/* Remove it from the DB */
		sql.delete_comp (mime);
	}

	public string get_info_using_plugins (string mime, string filename)
	{
		if (mime.Length>0) {
									
			if ( infoplugins.Contains(mime) ) {

				/* Do not extract files "*~" */
				if (filename.EndsWith("~")) {
				
					/* Do nothing */

				} else {

					InfoPluginItem plug = (InfoPluginItem) infoplugins[mime];
			
					string saveargs = plug.Args.Replace("{0}","\"{0}\"");

					/* Execute the external program to get the mime info */
					System.Diagnostics.Process proc = new System.Diagnostics.Process();
		
					proc.EnableRaisingEvents = false;
					proc.StartInfo.UseShellExecute = false;
					proc.StartInfo.FileName = plug.Program;
					proc.StartInfo.Arguments = String.Format (saveargs, filename);
					proc.StartInfo.RedirectStandardOutput = true;			
					proc.Start();
					proc.WaitForExit();

					/* Get raw output and format if using regular expressions */
					string rawout = proc.StandardOutput.ReadToEnd ();
										
					if (plug.RegexInput!="") {
				
						/* Change markups for pango */			
						//rawout = Regex.Replace( rawout, "&", "&amp;");
						//rawout = Regex.Replace( rawout, "<", "&lt;");
						//rawout = Regex.Replace( rawout, ">", "&gt;");
	
						try {
							if (plug.RegexCase) {
								rawout = Regex.Replace( rawout, plug.RegexInput, plug.RegexPattern);
							} else {
								rawout = Regex.Replace( rawout, plug.RegexInput, plug.RegexPattern, RegexOptions.IgnoreCase);
							}
						} catch {
							// Catch exceptions
						}
				
					}
				
					if (rawout.Length>0) {
						return(rawout);
					}
				
				}
			}

		}

		return("");
	}

	public int count_files (string dirname)
	{
		DirectoryInfo dir;
		int i;
		dir = new DirectoryInfo(dirname);
		int nfiles = 0;
		
		if (dir.Exists) {
		
			DirectoryInfo [] dirs = dir.GetDirectories();
			for (i=0; i<dirs.Length; i++) {
				nfiles += count_files (dirs[i].FullName);
			}
		  
            System.IO.FileInfo [] files = dir.GetFiles();		
		    nfiles += files.Length;
		    
		}
		return(nfiles);
	}

    	
	private Hashtable scan_cd_files (
	   string dirname, string parent_path, string parent_location, int cd_id, 
	   bool use_infoplugins, bool use_compplugins, bool compplugins_recursive, 
	   bool use_thumbs, bool packed,
	   ImportCDProgressWindow progress)
	{
		string fname, doing;
		int i,j;		
		DirectoryInfo dir = new DirectoryInfo (dirname);
		Hashtable filelist = new Hashtable ();
		
		if (dir.Exists) {
		
			DirectoryInfo [] dirs = dir.GetDirectories();
			for (i=0; i<dirs.Length; i++) {

				/* Create path of current file */
				string path = parent_path + i.ToString ();

				/* Create a file item with the directory */
				FileItem f = new FileItem (cd_id, path, dirs[i].Name, "", 0, "", 
					parent_location, FileItemType.Directory, packed);
                filelist.Add( f.Path, f);

						
				/* Recursive to subdirs */
				string p;
				if (parent_location.Length==1) {
					p = "/" + dirs[i].Name;
				} else {
					p = parent_location + "/" + dirs[i].Name;
				}
                Hashtable filelist2 = scan_cd_files (dirs[i].FullName, path + ":", 
                	p, 
                	cd_id, use_infoplugins, use_compplugins, compplugins_recursive,
                    use_thumbs, packed, progress);
                
                /* Add list to previous hashtable */
                if (filelist2==null) {
                    return(null);
                } else {
                    foreach (FileItem ff in filelist2.Values) {
                        filelist.Add( ff.Path, ff);
                    }
                }

			}

			System.IO.FileInfo [] files_unchecked = dir.GetFiles();
			Hashtable files = new Hashtable ();
			
			// Check if all files are ok */
			for (j=0; j<files_unchecked.Length; j++) {
				if (files_unchecked[j].Exists) {
					files.Add (j,files_unchecked[j]);
				}
			}			

			j = 0;
	        foreach (System.IO.FileInfo cfile in files.Values) {
		
				string path = "";
				string mime = "";
				string info = "";
				string uri  = "";

				path = parent_path + (i+j).ToString ();
				//mime = infomime.get_mime_type (files[j].FullName);
				//mime = Gnome.Vfs.MimeType.GetMimeTypeForUri("file://" + files[j].FullName);
			
				/* Notify external thread if needed */
				if (progress!=null) {

					if (cfile.Name.Length>30) {
						fname = "'" + cfile.Name.Substring (0,19) + "..." + 
								cfile.Name.Substring (cfile.Name.Length-8,8) + "'";
					} else {
						fname = "'" + cfile.Name + "'";
					}
					if (packed==true) {
						doing = Catalog.GetString ("Adding compressed file ");
					} else {
						doing = Catalog.GetString("Adding file ");
					}
    				progress.action =  doing + fname;
    				progress.actionnotify.WakeupMain ();

	                if (packed == false) {								
	    			    progress.pbar_steps++;
	    			    progress.stepnotify.WakeupMain ();
	    			}

    			}

				// Get MIME information from file
				uri = Gnome.Vfs.Uri.GetUriFromLocalPath (cfile.FullName);
				mime = Gnome.Vfs.MimeType.GetMimeTypeForUri(uri);
 
 				/* Console.WriteLine ("Getting mime from file '" + uri + "': mime = '" + mime + "'"); */
 
				/* Get info from plugins */
				if (use_infoplugins) {

					// Test if Get Mime type returned ok
					try {
						info = get_info_using_plugins (mime, cfile.FullName);
					} catch {
						//mime = "";
					}

				} else {
					info = "";
				}				
				
				/* Add file to local hashtable */
 				/* Console.WriteLine ("Adding file '" + cfile.Name + "'"); */
				FileItem f = new FileItem (cd_id, 
										   path, 
										   cfile.Name, 
										   info, 
										   (int)cfile.Length, 
										   mime, 
										   parent_location, 
										   FileItemType.File, 
										   packed);
                filelist.Add (f.Path,
                              f);                              
					
				/* Extract thumbnail for file */
				if (use_thumbs) {
					/* Console.WriteLine ("Importing thumbnail for " + cd_id.ToString() + "_" + f.Path); */
					import_thumbnail (uri,
					                  f.Mime,
					                  cd_id,
					                  f.ThumbKey());
				}
			

			    /* Extract if compressed */
			    if (use_compplugins) {
			    			    
	                if ( (!packed) || ( (packed) && (compplugins_recursive) ) ) {
                    
                    	if (CDCollect.db.compplugins.Contains(mime)) {

							/* Do not extract files "*~" */
							if (cfile.FullName.EndsWith("~")) {
								/* Do nothing */
							} else {
							
								/* Extract contents into a temporal folder */
    	       					if (progress!=null) {
									if (cfile.Name.Length>30) {
										fname = "'" + cfile.Name.Substring (0,19) + "..." + 
												cfile.Name.Substring (cfile.Name.Length-8,8) + "'";
									} else {
										fname = "'" + cfile.Name + "'";
									}
    	           					progress.action = Catalog.GetString ("Extracting ") + fname;
		               				progress.actionnotify.WakeupMain ();
    		           			}
                       
								/* Create temporal directory */
								string tmpdir = System.IO.Path.GetTempFileName();
								System.IO.File.Delete (tmpdir);
								System.IO.Directory.CreateDirectory (tmpdir);

								/* Extract contents */
								CompPluginItem c = (CompPluginItem) compplugins[mime];
								string saveargs = c.Args.Replace("{0}", "\"{0}\"").Replace("{1}","\"{1}\"");

								System.Diagnostics.Process proc = new System.Diagnostics.Process();
								proc.EnableRaisingEvents = false;
								proc.StartInfo.FileName = c.Program;
								string args = String.Format(saveargs, cfile.FullName, tmpdir);
								proc.StartInfo.Arguments = args; 
								
								Trace.WriteLine (c.Program + " " + args);

								proc.Start();
								proc.WaitForExit();
			            			            
								/* Recursive add to database (with packed on) */                        
								Hashtable filelist2 = scan_cd_files (tmpdir, path + ":", 
									parent_location,
									cd_id, use_infoplugins, use_compplugins, 
									compplugins_recursive, use_thumbs, true,
    	                        	progress);
                        
								/* Add list to previous hashtable */
								if (filelist2!=null) {
									foreach (FileItem ff in filelist2.Values) {
										filelist.Add( ff.Path, ff);
									}
								}
                                                
								/* Remove created directory */
								// FIXME: hack to fix a bug in "System.IO.Directory.Delete"
								// when deleting symbolic links 
								delete_directory_files (tmpdir);
								try {
									System.IO.Directory.Delete (tmpdir, true);
								} catch {
									// "Error trying to delete temporal folder, garbage will be left over"
									Console.WriteLine ( String.Format (Catalog.GetString ("Warning, temporal folder '{0}' could not be deleted"), tmpdir)); 
								}
								
							} /* not ends with "~" */
							
						} /* mime is compressed */
                    } /* is not packed or packed but recursive */
                } /* use_compplugins */


				j++;
				
			} /* for each file */

			return(filelist);

		} else {
		
			return(null);
		
		}
		
	}
	
	
	private void delete_directory_files (string dirname) 
	{
			
		DirectoryInfo dir = new DirectoryInfo (dirname);
		
		
		if (dir.Exists) {
		
			// FIXME: This will only work in linux!!			
			System.Diagnostics.Process proc = new System.Diagnostics.Process();
			proc.EnableRaisingEvents = false;
			proc.StartInfo.FileName = "rm";
			proc.StartInfo.Arguments = " -rf \"" + dirname + "\"/*"; 
			proc.Start();
			proc.WaitForExit();
	
		}
	
	}


	public ArrayList find_files (	bool doname, string nametype, string namefind, bool namecase,
									bool dosize, string sizetype, int sizefind)
	{
		return (sql.find_files (doname, nametype, namefind, namecase, dosize, sizetype, sizefind));
	}

	public Hashtable read_files (int cd_key)
	{
		return (sql.read_files (cd_key));
	}

	public Hashtable read_cds ()
	{
		return (sql.read_cds ());
	}

	public Hashtable read_cds (ArrayList files)
	{
		return (sql.read_cds (files));
	}

	public CDItem read_cds (int key)
	{
		return (sql.read_cds (key));
	}

    public int import_cd_step1 ( 	string mount,
    								string name, 
    								string keywords,
    								DateTime date, 
    								Pixbuf icon, 
    								bool use_infoplugins, 
    								bool use_compplugins, 
    								bool compplugins_recursive,
	                    			ImportCDProgressWindow progress,
	                    			out CDItem cd)
	{
		cd = null;
		
		/* Check if directory exists and has 0 files */
		DirectoryInfo dir = new DirectoryInfo (mount);
		if (!dir.Exists) {
			return(-1);
		}
		DirectoryInfo [] mntdirs = dir.GetDirectories();
		System.IO.FileInfo [] mntfiles = dir.GetFiles();
		if ( (mntdirs.Length + mntfiles.Length) == 0) {
			return(-2);
		}

		/* Write CD to database (in order to get the key) */
		cd = new CDItem (-1, name, keywords, date, icon);
		cd.Key = sql.write_cd (cd.Name, cd.Keywords, cd.Date, cd.Icon.get_icon_serial()); 

		return (cd.Key);
	}
	
    public int import_cd_step2 (CDItem cd, 
                                string mount, 
                                string name, 
                                string keywords, 
                                DateTime date, 
                                Pixbuf icon, 
    							bool use_infoplugins, 
    							bool use_compplugins, 
    							bool compplugins_recursive,
	                    		bool use_thumbs, 
	                    		ImportCDProgressWindow progress)
	{
		Hashtable files;
		
		/* Read files */
		files = scan_cd_files(mount, "", "/", cd.Key, use_infoplugins, use_compplugins, 
				compplugins_recursive, use_thumbs, false, progress);
              
		if (files == null) {
			sql.delete_cds (cd.Key);
			return(-3);
		}

		/* Notify of writing files to DB */
		progress.action = Catalog.GetString ("Writing files to Database.");
		progress.actionnotify.WakeupMain ();
		
	   	/* Write files */	   
	   	sql.write_files (cd.Key, files);
	   	
		/* Notify of writing files to DB */
		progress.action = Catalog.GetString ("Finished.");
		progress.actionnotify.WakeupMain ();

	   	/* Add cd to cds hashtable */
	   	if (cds.Contains(cd.Key)) {
	   		cds.Remove(cd.Key);
	   	}
	   	cds.Add (cd.Key, cd);
	   	
	   	return(cd.Key);
	   	
    }
    
    
	public void delete_cd_and_files (int cd_key)
	{
		sql.delete_cds (cd_key);
		sql.delete_files (cd_key);
		sql.delete_thumbnails (cd_key);
		
		if (cds.Contains(cd_key)) {
			cds.Remove(cd_key);
		}
	}	

	public void update_cd (CDItem cd)
	{
		/* Update cd in hashtable */
		if ( cds.Contains (cd.Key) ) {
			cds.Remove(cd.Key);
		}
		cds.Add (cd.Key, cd);
	
	    sql.update_cd (cd);
	}


	public Pixbuf read_thumbnail (int CDKey, string thumbkey)
	{
		return (sql.read_thumbnail (CDKey,thumbkey));
	}
	
	public void write_thumbnail (Pixbuf thumb,int CDKey,string thumbkey)
	{
		if (thumb!=null) {
			sql.write_thumbnail (thumb,CDKey,thumbkey);
		}
	}

	public void import_thumbnail (string uri,string mime,int CDKey,string thumbskey)
	{
		Pixbuf thumb = null;
				
		/* generate thumbnail */
		Trace.WriteLine ("Generating thumbnail for mime '" + mime + "' for file '" + uri + "'");
		
		Gnome.ThumbnailFactory thumbnailfactory = new Gnome.ThumbnailFactory (Gnome.ThumbnailSize.Normal);
		thumb = thumbnailfactory.GenerateThumbnail (uri,mime); 
		
		/* write it to disk */
		write_thumbnail (thumb, CDKey, thumbskey);
	}

	public void delete_thumbnails (int CDKey)
	{
		sql.delete_thumbnails (CDKey);
	}


	public void update_file (FileItem f,Pixbuf thumbnail)
	{
		if (thumbnail==null) {
			sql.delete_thumbnail (f.CD_Key,f.ThumbKey());
		} else {
			write_thumbnail (thumbnail, f.CD_Key, f.ThumbKey());
		}	
	    sql.update_file (f);
	}


	public string dump_cds () 
	{
		string dump = "";
		foreach (CDItem cd in CDCollect.db.cds.Values) {
			Hashtable files = read_files (cd.Key);
			dump = dump + "cd = " + cd.Key + "; title = '" + cd.Name + "'; files = " + files.Count + "\n";
		}
		return (dump);
	}

}

